"""
Package for working with JSON-format configuration files.
"""
from ._JSONObject import JSONObject
from ._StrictJSONObject import StrictJSONObject
from ._typing import (
    Absent,
    OptionallyPresent,
    PropertyValueType
)
