"""
Package specifying interfaces for objects/classes which can perform JSON validation.
"""
from ._BasicSchemaValidator import BasicSchemaValidator
from ._JSONValidator import JSONValidator
from ._StaticJSONValidator import StaticJSONValidator
